from enum import Enum

__all__ = ["Distribution"]


class Distribution(str, Enum):
    """
    Names of probability distributions.
    """

    NORMAL = "Normal"
    LOGNORMAL = "Lognormal"
    EXPONENTIAL = "Exponential"
    UNIFORM = "Uniform"
    PARETO = "Pareto"
    GAMMA = "Gamma"
    BERNOULLI = "Bernoulli"
    CATEGORICAL = "Categorical"
    GEOMETRIC = "Geometric"
    POISSON = "Poisson"



__version__ = "0.1.16"
__author__ = "Mehdi Samsami"


from ._base import ContinuousDistMixin, DiscreteDistMixin
from ._enums import Distribution
from .gnb import GeneralNB
from .gwnb import GaussianWNB


__all__ = [
    "GeneralNB",
    "GaussianWNB",
    "Distribution",
    "ContinuousDistMixin",
    "DiscreteDistMixin",
]



import numpy as np

import pytest
from sklearn.utils.estimator_checks import check_estimator
from sklearn.base import is_classifier
from sklearn.utils._testing import assert_array_equal, assert_array_almost_equal

from wnb import GaussianWNB

# Data is just 6 separable points in the plane
X = np.array([[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]])
y = np.array([1, 1, 1, 2, 2, 2])


@pytest.fixture
def global_random_seed():
    return np.random.randint(0, 1000)


def get_random_normal_x_binary_y(global_random_seed):
    # A bit more random tests
    rng = np.random.RandomState(global_random_seed)
    X1 = rng.normal(size=(10, 3))
    y1 = (rng.normal(size=10) > 0).astype(int)
    return X1, y1


def test_gwnb():
    """Binary Gaussian MLD-WNB classification

    Checks that GaussianWNB implements fit and predict and returns correct values for a simple toy dataset.
    """
    clf = GaussianWNB()
    y_pred = clf.fit(X, y).predict(X)
    assert_array_equal(y_pred, y)

    y_pred_proba = clf.predict_proba(X)
    y_pred_log_proba = clf.predict_log_proba(X)
    assert_array_almost_equal(np.log(y_pred_proba), y_pred_log_proba, 8)


def test_gwnb_estimator():
    """
    Test whether GaussianWNB estimator adheres to scikit-learn conventions.
    """
    check_estimator(GaussianWNB())
    assert is_classifier(GaussianWNB)


def test_gwnb_prior(global_random_seed):
    """
    Test whether class priors are properly set.
    """
    clf = GaussianWNB().fit(X, y)
    assert_array_almost_equal(np.array([3, 3]) / 6.0, clf.class_prior_, 8)

    X1, y1 = get_random_normal_x_binary_y(global_random_seed)
    clf = GaussianWNB().fit(X1, y1)

    # Check that the class priors sum to 1
    assert_array_almost_equal(clf.class_prior_.sum(), 1)


def test_gwnb_neg_priors():
    """
    Test whether an error is raised in case of negative priors.
    """
    clf = GaussianWNB(priors=np.array([-1.0, 2.0]))

    msg = "Priors must be non-negative"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_priors():
    """
    Test whether the class priors override is properly used.
    """
    clf = GaussianWNB(priors=np.array([0.3, 0.7])).fit(X, y)
    assert_array_almost_equal(
        clf.predict_proba([[-0.1, -0.1]]),
        np.array([[0.82357095, 0.17642905]]),
        8,
    )
    assert_array_almost_equal(clf.class_prior_, np.array([0.3, 0.7]))


def test_gwnb_wrong_nb_priors():
    """
    Test whether an error is raised if the number of priors is different from the number of classes.
    """
    clf = GaussianWNB(priors=np.array([0.25, 0.25, 0.25, 0.25]))

    msg = "Number of priors must match the number of classes"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_prior_greater_one():
    """
    Test if an error is raised if the sum of priors greater than one.
    """
    clf = GaussianWNB(priors=np.array([2.0, 1.0]))

    msg = "The sum of the priors should be 1"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_prior_large_bias():
    """
    Test if good prediction when class priors favor largely one class.
    """
    clf = GaussianWNB(priors=np.array([0.01, 0.99]))
    clf.fit(X, y)
    assert clf.predict(np.array([[-0.1, -0.1]])) == np.array([2])


def test_gwnb_non_binary():
    """
    Test if an error is raised when given non-binary targets.
    """
    X_ = np.array(
        [
            [-1, -1],
            [-2, -1],
            [-3, -2],
            [-4, -5],
            [-5, -4],
            [1, 1],
            [2, 1],
            [3, 2],
            [4, 4],
            [5, 5],
        ]
    )
    y_ = np.array([1, 2, 3, 4, 4, 3, 2, 1, 1, 2])
    clf = GaussianWNB()

    msg = "Unknown label type: non-binary"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X_, y_)


def test_gwnb_wrong_error_weights():
    """
    Test whether an error is raised in case error weight is not a square array of size (n_classes, n_classes).
    """
    clf = GaussianWNB(error_weights=np.array([[1, 2, 0], [0, -2, -3]]))

    msg = "The shape of error weights matrix does not match the number of classes"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)

    clf = GaussianWNB(error_weights=np.array([[1, 2, 0], [0, -2, -3], [-2, 1, 1.5]]))
    msg = "The shape of error weights matrix does not match the number of classes"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_wrong_penalty():
    """
    Test whether an error is raised in case regularization penalty is not supported.
    """
    clf = GaussianWNB(penalty="dropout")

    msg = "Regularization type must be either 'l1' or 'l2'"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_neg_C():
    """
    Test whether an error is raised in case of negative regularization parameter, C.
    """
    clf = GaussianWNB(C=-0.6)

    msg = "Regularization parameter must be positive"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_neg_max_iter():
    """
    Test whether an error is raised in case number of iteration is negative.
    """
    clf = GaussianWNB(max_iter=-10)

    msg = "Maximum number of iteration must be a positive integer"
    with pytest.raises(ValueError, match=msg):
        clf.fit(X, y)


def test_gwnb_no_cost_hist():
    """
    Test whether cost_hist_ is None if learning_hist is not enabled.
    """
    clf = GaussianWNB(max_iter=10)
    clf.fit(X, y)
    assert clf.cost_hist_ is None



from abc import ABCMeta
import numbers
from typing import Union, Optional, Sequence
import warnings

import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.special import logsumexp

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils import as_float_array, check_array, deprecated
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import type_of_target

__all__ = [
    "GaussianWNB",
]


class GaussianWNB(ClassifierMixin, BaseEstimator, metaclass=ABCMeta):
    """
    Binary Gaussian Minimum Log-likelihood Difference Weighted Naive Bayes (MLD-WNB) Classifier
    """

    def __init__(
        self,
        *,
        priors: Optional[Union[Sequence[float], np.ndarray]] = None,
        error_weights: Optional[np.ndarray] = None,
        max_iter: int = 25,
        step_size: float = 1e-4,
        penalty: str = "l2",
        C: float = 1.0,
        learning_hist: bool = False
    ) -> None:
        """Initializes an instance of the GaussianWNB class.

        Args:
            priors (Optional[Union[Sequence[float], np.ndarray]]): Prior probabilities. Defaults to None.
            error_weights (Optional[np.ndarray]): Matrix of error weights (n_classes * n_classes). Defaults to None.
            max_iter (int): Maximum number of gradient descent iterations. Defaults to 25.
            step_size (float): Step size of weight update (i.e., learning rate). Defaults to 1e-4.
            penalty (str): Regularization term; must be either 'l1' or 'l2'. Defaults to 'l2'.
            C (float): Regularization strength; must be a positive float. Defaults to 1.0.
            learning_hist (bool): Whether to record the learning history, i.e., the value of cost function in each
                                  learning iteration.

        """
        self.priors = priors
        self.error_weights = error_weights
        self.max_iter = max_iter
        self.step_size = step_size
        self.penalty = penalty
        self.C = C
        self.learning_hist = learning_hist

    def _more_tags(self):
        return {"binary_only": True, "requires_y": True}

    def _check_inputs(self, X, y):
        # Check that the dataset has only two unique labels
        if type_of_target(y) != "binary":
            warnings.warn(
                "This version of MLD-WNB only supports binary classification."
            )
            raise ValueError("Unknown label type: non-binary")

        # Check if only one class is present in label vector
        if self.n_classes_ == 1:
            raise ValueError("Classifier can't train when only one class is present.")

        X = check_array(
            array=X,
            accept_sparse=False,
            accept_large_sparse=False,
            dtype="numeric",
            force_all_finite=True,
            ensure_2d=True,
            ensure_min_samples=1,
            ensure_min_features=1,
            estimator=self,
        )

        # Check if X contains complex values
        if np.iscomplex(X).any() or np.iscomplex(y).any():
            raise ValueError("Complex data not supported")

        # Check that the number of samples and labels are compatible
        if self.__n_samples != y.shape[0]:
            raise ValueError(
                "X.shape[0]=%d and y.shape[0]=%d are incompatible."
                % (X.shape[0], y.shape[0])
            )

        if self.priors is not None:
            # Check that the provided priors match the number of classes
            if len(self.priors) != self.n_classes_:
                raise ValueError("Number of priors must match the number of classes.")
            # Check that the sum of priors is 1
            if not np.isclose(self.priors.sum(), 1.0):
                raise ValueError("The sum of the priors should be 1.")
            # Check that the priors are non-negative
            if (self.priors < 0).any():
                raise ValueError("Priors must be non-negative.")

        if self.error_weights is not None:
            # Check that the size of error weights matrix matches number of classes
            if self.error_weights.shape != (self.n_classes_, self.n_classes_):
                raise ValueError(
                    "The shape of error weights matrix does not match the number of classes, "
                    "must be (n_classes, n_classes)."
                )

        # Check that the regularization type is either 'l1' or 'l2'
        if self.penalty not in ["l1", "l2"]:
            raise ValueError("Regularization type must be either 'l1' or 'l2'.")

        # Check that the regularization parameter is a positive integer
        if not isinstance(self.C, numbers.Number) or self.C < 0:
            raise ValueError(
                "Regularization parameter must be positive; got (C=%r)" % self.C
            )

        # Check that the maximum number of iterations is a positive integer
        if not isinstance(self.max_iter, numbers.Number) or self.max_iter < 0:
            raise ValueError(
                "Maximum number of iteration must be a positive integer; got (max_iter=%r)."
                % self.max_iter
            )

    def _prepare_X_y(self, X=None, y=None, from_fit=False):
        if from_fit and y is None:
            raise ValueError("requires y to be passed, but the target y is None.")

        if X is not None:
            # Convert to NumPy array if X is Pandas DataFrame
            if isinstance(X, pd.DataFrame):
                X = X.values
            X = as_float_array(X)

        if y is not None:
            # Convert to a NumPy array
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
            else:
                y = np.array(y)

            # Warning in case of y being 2d
            if y.ndim > 1:
                warnings.warn(
                    "A column-vector y was passed when a 1d array was expected.",
                    DataConversionWarning,
                )

            y = y.flatten()

        output = tuple(item for item in [X, y] if item is not None)
        output = output[0] if len(output) == 1 else output
        return output

    def _prepare_parameters(self, X, y):
        # Calculate mean and standard deviation of features for each class
        for c in range(self.n_classes_):
            self.mu_[:, c] = np.mean(
                X[y == c, :], axis=0
            )  # Calculate mean of features for class c
            self.std_[:, c] = np.std(
                X[y == c, :], axis=0
            )  # Calculate std of features for class c

        # Update if no priors is provided
        if self.priors is None:
            _, class_count_ = np.unique(y, return_counts=True)
            self.class_prior_ = (
                class_count_ / class_count_.sum()
            )  # Calculate empirical prior probabilities
        else:
            self.class_prior_ = self.priors

        # Convert to NumPy array if input priors is in a list
        if type(self.class_prior_) is list:
            self.class_prior_ = np.array(self.class_prior_)

        # Update if no error weights is provided
        if self.error_weights is None:
            self.error_weights_ = np.array([[0, 1], [-1, 0]])
        else:
            self.error_weights_ = self.error_weights

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame, pd.Series],
    ):
        """Fits Gaussian Binary MLD-WNB to X and y.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features).
                                                 Training vectors, where `n_samples` is the number of samples
                                                 and `n_features` is the number of features.
            y (Union[np.ndarray, pd.DataFrame, pd.Series]): Array-like of shape (n_samples,). Target values.

        Returns:
            self: The instance itself.
        """
        X, y = self._prepare_X_y(X, y, from_fit=True)

        self.classes_, y_ = np.unique(
            y, return_inverse=True
        )  # Unique class labels and their indices
        self.n_classes_ = len(self.classes_)  # Number of classes
        (
            self.__n_samples,
            self.n_features_in_,
        ) = X.shape  # Number of samples and features

        self._check_inputs(X, y)
        y = y_

        self.mu_ = np.zeros(
            (self.n_features_in_, self.n_classes_)
        )  # Mean of features (n_features x 1)
        self.std_ = np.zeros(
            (self.n_features_in_, self.n_classes_)
        )  # Standard deviation of features (n_features x 1)
        self.coef_ = np.ones(
            (self.n_features_in_,)
        )  # WNB coefficients (n_features x 1)
        self.cost_hist_ = np.array(
            [np.nan for _ in range(self.max_iter)]
        )  # To store cost value in each iteration

        self._prepare_parameters(X, y)

        # Learn the weights using gradient descent
        self.n_iter_ = 0
        for self.n_iter_ in range(self.max_iter):
            # Predict on X
            y_hat = self.__predict(X)

            # Calculate cost
            self.cost_hist_[self.n_iter_], _lambda = self._calculate_cost(
                X, y, y_hat, self.learning_hist
            )

            # Calculate gradients (most time-consuming)
            _grad = self._calculate_grad(X, _lambda)

            # Add regularization
            if self.penalty == "l1":
                _grad += self.C * np.sign(self.coef_)
            elif self.penalty == "l2":
                _grad += 2 * self.C * self.coef_

            # Update weights
            self.coef_ = self.coef_ - self.step_size * _grad

        self.cost_hist_ = None if not self.learning_hist else self.cost_hist_

        return self

    def _calculate_cost(self, X, y, y_hat, learning_hist):
        _lambda = [self.error_weights_[y[i], y_hat[i]] for i in range(self.__n_samples)]

        if learning_hist:
            # Calculate cost
            _cost = 0
            for i in range(self.__n_samples):
                _sum = np.log(self.class_prior_[1] / self.class_prior_[0])
                x = X[i, :]
                for j in range(self.n_features_in_):
                    _sum += self.coef_[j] * (
                        np.log(1e-20 + norm.pdf(x[j], self.mu_[j, 1], self.std_[j, 1]))
                        - np.log(
                            1e-20 + norm.pdf(x[j], self.mu_[j, 0], self.std_[j, 0])
                        )
                    )
                _cost += _lambda[i] * _sum
        else:
            _cost = np.nan

        return _cost, _lambda

    def _calculate_grad(self, X, _lambda):
        _grad = np.repeat(
            np.log(self.std_[:, 0] / self.std_[:, 1]).reshape(1, -1),
            self.__n_samples,
            axis=0,
        )
        _grad += (
            0.5
            * (
                (X - np.repeat(self.mu_[:, 0].reshape(1, -1), self.__n_samples, axis=0))
                / (np.repeat(self.std_[:, 0].reshape(1, -1), self.__n_samples, axis=0))
            )
            ** 2
        )
        _grad -= (
            0.5
            * (
                (X - np.repeat(self.mu_[:, 1].reshape(1, -1), self.__n_samples, axis=0))
                / (np.repeat(self.std_[:, 1].reshape(1, -1), self.__n_samples, axis=0))
            )
            ** 2
        )
        _grad *= np.transpose(
            np.repeat(np.array(_lambda).reshape(1, -1), self.n_features_in_, axis=0)
        )
        _grad = np.sum(_grad, axis=0)

        return _grad

    @deprecated()
    def _calculate_grad_slow(self, X, _lambda):
        _grad = np.zeros((self.n_features_in_,))
        for i in range(self.__n_samples):
            x = X[i, :]
            _log_p = np.array(
                [
                    np.log(self.std_[j, 0] / self.std_[j, 1])
                    + 0.5 * ((x[j] - self.mu_[j, 0]) / self.std_[j, 0]) ** 2
                    - 0.5 * ((x[j] - self.mu_[j, 1]) / self.std_[j, 1]) ** 2
                    for j in range(self.n_features_in_)
                ]
            )
            _grad += _lambda[i] * _log_p
        return _grad

    def __predict(self, X):
        p_hat = self.predict_log_proba(X)
        return np.argmax(p_hat, axis=1)

    def predict(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Performs classification on an array of test vectors X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: ndarray of shape (n_samples,). Predicted target values for X.
        """
        p_hat = self.predict_log_proba(X)
        y_hat = self.classes_[np.argmax(p_hat, axis=1)]
        return y_hat

    def predict_log_proba(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Returns log-probability estimates for the test vector X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: Array-like of shape (n_samples, n_classes).
                        The log-probability of the samples for each class in the model.
                        The columns correspond to the classes in sorted order, as they appear in the attribute `classes_`.
        """
        # Check is fit had been called
        check_is_fitted(self)

        # Input validation
        X = check_array(
            array=X, accept_large_sparse=False, force_all_finite=True, estimator=self
        )

        # Check if the number of input features matches the data seen during fit
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Expected input with %d features, got %d instead."
                % (self.n_features_in_, X.shape[1])
            )

        n_samples = X.shape[0]

        X = self._prepare_X_y(X=X)

        log_priors = np.tile(np.log(self.class_prior_), (n_samples, 1))
        w_reshaped = np.tile(self.coef_.reshape(-1, 1), (1, self.n_classes_))
        term1 = np.sum(np.multiply(w_reshaped, -np.log(np.sqrt(2 * np.pi) * self.std_)))
        var_inv = np.multiply(w_reshaped, 1.0 / np.multiply(self.std_, self.std_))
        mu_by_var = np.multiply(self.mu_, var_inv)
        term2 = -0.5 * (
            np.matmul(np.multiply(X, X), var_inv)
            - 2.0 * np.matmul(X, mu_by_var)
            + np.sum(self.mu_.conj() * mu_by_var, axis=0)
        )
        log_proba = log_priors + term1 + term2

        log_proba -= np.transpose(
            np.repeat(
                logsumexp(log_proba, axis=1).reshape(1, -1), self.n_classes_, axis=0
            )
        )
        return log_proba

    def predict_proba(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Returns probability estimates for the test vector X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: Array-like of shape (n_samples, n_classes).
                        The probability of the samples for each class in the model.
                        The columns correspond to the classes in sorted order, as they appear in the attribute `classes_`.
        """
        return np.exp(self.predict_log_proba(X))



from abc import ABCMeta
from typing import Union, Optional, Sequence, Type
import warnings

import numpy as np
import pandas as pd
from scipy.special import logsumexp

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils import check_array, as_float_array
from sklearn.utils.validation import check_is_fitted

from ._base import ContinuousDistMixin, DiscreteDistMixin
from ._enums import Distribution
from .dist import AllDistributions, NonNumericDistributions

__all__ = [
    "GeneralNB",
]


class GeneralNB(ClassifierMixin, BaseEstimator, metaclass=ABCMeta):
    """
    A general Naive Bayes classifier that allows you to specify the likelihood distribution for each feature.
    """

    def __init__(
        self,
        *,
        priors: Optional[Union[Sequence[float], np.ndarray]] = None,
        distributions: Optional[
            Sequence[
                Union[
                    str,
                    Distribution,
                    Type[ContinuousDistMixin],
                    Type[DiscreteDistMixin],
                ]
            ]
        ] = None,
        alpha: float = 1e-10,
    ) -> None:
        """Initializes an instance of the GeneralNB class.

        Args:
            priors (Optional[Union[list, np.ndarray]]): Prior probabilities. Defaults to None.
            distributions: Probability distributions to be used for features' likelihoods. A sequence with same length
                           of the number of features. If not specified, all likelihood will be considered Gaussian.
                           Defaults to None.
            alpha (float): Additive (Laplace/Lidstone) smoothing parameter (set alpha=0 for no smoothing). Defaults to 1e-10.

        """
        self.priors = priors
        self.distributions = distributions
        self.alpha = alpha

    def _more_tags(self):
        return {"multilabel": True, "requires_y": True}

    def _get_distributions(self):
        try:
            if self.distributions_ is not None:
                return self.distributions_
        except Exception:
            return self.distributions or []

    def _check_inputs(self, X, y):
        # Check if only one class is present in label vector
        if self.n_classes_ == 1:
            raise ValueError("Classifier can't train when only one class is present")

        X = check_array(
            array=X,
            accept_sparse=False,
            accept_large_sparse=False,
            dtype=None
            if any(d in self._get_distributions() for d in NonNumericDistributions)
            else "numeric",
            force_all_finite=True,
            ensure_2d=True,
            ensure_min_samples=1,
            ensure_min_features=1,
            estimator=self,
        )

        # Check if X contains complex values
        if np.iscomplex(X).any() or np.iscomplex(y).any():
            raise ValueError("Complex data not supported")

        # Check that the number of samples and labels are compatible
        if self.__n_samples != y.shape[0]:
            raise ValueError(
                "X.shape[0]=%d and y.shape[0]=%d are incompatible."
                % (X.shape[0], y.shape[0])
            )

    def _prepare_X_y(self, X=None, y=None):
        if X is not None:
            # Convert to NumPy array if X is Pandas DataFrame
            if isinstance(X, pd.DataFrame):
                X = X.values
            X = (
                X
                if any(d in self._get_distributions() for d in NonNumericDistributions)
                else as_float_array(X)
            )

        if y is not None:
            # Convert to a NumPy array
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
            else:
                y = np.array(y)

            # Warning in case of y being 2d
            if y.ndim > 1:
                warnings.warn(
                    "A column-vector y was passed when a 1d array was expected.",
                    DataConversionWarning,
                )

            y = y.flatten()

        output = tuple(item for item in [X, y] if item is not None)
        output = output[0] if len(output) == 1 else output
        return output

    def _prepare_parameters(self, X, y):
        # Set priors if not specified
        if self.priors is None:
            _, class_count_ = np.unique(y, return_counts=True)
            self.class_prior_ = (
                class_count_ / class_count_.sum()
            )  # Calculate empirical prior probabilities

        else:
            # Check that the provided priors match the number of classes
            if len(self.priors) != self.n_classes_:
                raise ValueError("Number of priors must match the number of classes.")
            # Check that the sum of priors is 1
            if not np.isclose(self.priors.sum(), 1.0):
                raise ValueError("The sum of the priors should be 1.")
            # Check that the priors are non-negative
            if (self.priors < 0).any():
                raise ValueError("Priors must be non-negative.")

            self.class_prior_ = self.priors

        # Convert to NumPy array if input priors is in a list
        if type(self.class_prior_) is list:
            self.class_prior_ = np.array(self.class_prior_)

        # Set distributions if not specified
        if self.distributions is None:
            self.distributions_ = [Distribution.NORMAL] * self.n_features_in_

        else:
            # Check if the number of distributions matches the number of features
            if len(self.distributions) != self.n_features_in_:
                raise ValueError(
                    "Number of specified distributions must match the number of features."
                    f"({len(self.distributions)} != {self.n_features_in_})"
                )

            # Check that all specified distributions are supported
            for dist in self.distributions:
                if not (
                    isinstance(dist, Distribution)
                    or dist in Distribution.__members__.values()
                    or (hasattr(dist, "from_data") and hasattr(dist, "__call__"))
                ):
                    raise ValueError(f"Distribution '{dist}' is not supported.")

            self.distributions_ = self.distributions

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame, pd.Series],
    ):
        """Fits general Naive Bayes classifier to X and y.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features).
                                                 Training vectors, where `n_samples` is the number of samples
                                                 and `n_features` is the number of features.
            y (Union[np.ndarray, pd.DataFrame, pd.Series]): Array-like of shape (n_samples,). Target values.

        Returns:
            self: The instance itself.
        """
        self.classes_, y_ = np.unique(
            y, return_inverse=True
        )  # Unique class labels and their indices
        self.n_classes_ = len(self.classes_)  # Number of classes

        X, y = self._prepare_X_y(X, y)

        (
            self.__n_samples,
            self.n_features_in_,
        ) = X.shape  # Number of samples and features

        self._check_inputs(X, y)

        y = y_
        self._prepare_parameters(X, y)

        self.likelihood_params_ = {
            c: [
                AllDistributions[self.distributions_[i]].from_data(
                    X[y == c, i], alpha=self.alpha
                )
                if isinstance(self.distributions_[i], Distribution)
                or self.distributions_[i] in Distribution.__members__.values()
                else self.distributions_[i].from_data(X[y == c, i], alpha=self.alpha)
                for i in range(self.n_features_in_)
            ]
            for c in range(self.n_classes_)
        }

        return self

    def predict(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Performs classification on an array of test vectors X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: ndarray of shape (n_samples,). Predicted target values for X.
        """
        p_hat = self.predict_log_proba(X)
        y_hat = self.classes_[np.argmax(p_hat, axis=1)]
        return y_hat

    def predict_log_proba(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Returns log-probability estimates for the test vector X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: Array-like of shape (n_samples, n_classes).
                        The log-probability of the samples for each class in the model.
                        The columns correspond to the classes in sorted order, as they appear in the attribute `classes_`.
        """
        # Check is fit had been called
        check_is_fitted(self)

        # Input validation
        X = check_array(
            array=X,
            accept_large_sparse=False,
            force_all_finite=True,
            dtype=None
            if any(d in self._get_distributions() for d in NonNumericDistributions)
            else "numeric",
            estimator=self,
        )

        # Check if the number of input features matches the data seen during fit
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Expected input with %d features, got %d instead."
                % (self.n_features_in_, X.shape[1])
            )

        n_samples = X.shape[0]

        X = self._prepare_X_y(X=X)

        log_joint = np.zeros((n_samples, self.n_classes_))
        for c in range(self.n_classes_):
            log_joint[:, c] = np.log(self.class_prior_[c]) + np.sum(
                [
                    np.log(likelihood(X[:, i]))
                    for i, likelihood in enumerate(self.likelihood_params_[c])
                ],
                axis=0,
            )

        log_proba = log_joint - np.transpose(
            np.repeat(
                logsumexp(log_joint, axis=1).reshape(1, -1), self.n_classes_, axis=0
            )
        )
        return log_proba

    def predict_proba(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """Returns probability estimates for the test vector X.

        Args:
            X (Union[np.ndarray, pd.DataFrame]): Array-like of shape (n_samples, n_features). The input samples.

        Returns:
            np.ndarray: Array-like of shape (n_samples, n_classes).
                        The probability of the samples for each class in the model.
                        The columns correspond to the classes in sorted order, as they appear in the attribute `classes_`.
        """
        return np.exp(self.predict_log_proba(X))



import codecs
from os import path
from setuptools import setup, find_packages


setup(
    name="wnb",
    version="0.1.16",
    description="Python library for the implementations of general and weighted naive Bayes (WNB) classifiers.",
    keywords=["python", "bayes", "naivebayes", "classifier", "probabilistic"],
    author="Mehdi Samsami",
    author_email="mehdisamsami@live.com",
    url="https://github.com/msamsami/weighted-naive-bayes",
    long_description=codecs.open(
        path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8"
    ).read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.7",
    install_requires=["pandas==1.4.1", "scikit-learn>=1.0.2"],
    extras_require={"dev": ["pytest==7.3.1", "black>=23.9.0"]},
)



from abc import ABCMeta
from functools import wraps
import inspect
from numbers import Number
from typing import List, Tuple, Union
import warnings

import numpy as np

from ._enums import Distribution

__all__ = ["ContinuousDistMixin", "DiscreteDistMixin"]


def vectorize(otypes=None, excluded=None, signature=None):
    """
    Numpy vectorization wrapper that works with class methods.
    """

    def decorator(func):
        vectorized = np.vectorize(
            func, otypes=otypes, excluded=excluded, signature=signature
        )

        @wraps(func)
        def wrapper(*args):
            return vectorized(*args)

        return wrapper

    return decorator


class DistMixin(metaclass=ABCMeta):
    """
    Mixin class for probability distributions in wnb.
    """

    name: Union[str, Distribution] = None
    _support: Union[List[float], Tuple[float, float]] = None

    @classmethod
    def from_data(cls, data, **kwargs):
        """Creates an instance of the class from given data. Distribution parameters will be estimated from data.

        Args:
            data: Input data from which distribution parameters will be estimated.

        Returns:
            self: An instance of the class.
        """
        pass

    @classmethod
    def _get_param_names(cls):
        """
        Gets parameter names for the distribution instance.
        """
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            return []

        init_signature = inspect.signature(init)
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]

        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "wnb estimators should always "
                    "specify their parameters in the signature "
                    "of their __init__ (no varargs). "
                    "%s with constructor %s doesn't "
                    "follow this convention." % (cls, init_signature)
                )

        return sorted([p.name for p in parameters])

    def get_params(self) -> dict:
        """Gets parameters for this distribution instance.

        Returns:
            dict: Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            out[key] = value
        return out

    @property
    def support(self) -> Union[List[float], Tuple[float, float]]:
        """Returns the support of the probability distribution.

        If support is a list, it represents a limited number of discrete values.
        If it is a tuple, it indicates a limited or unlimited range of continuous values.
        """
        return self._support

    def _check_support(self, x):
        if (isinstance(self.support, list) and x not in self.support) or (
            isinstance(self.support, tuple)
            and (x < self.support[0] or x > self.support[1])
        ):
            warnings.warn(
                "Value doesn't lie within the support of the distribution",
                RuntimeWarning,
            )

    def __repr__(self) -> str:
        return "".join(
            [
                "<",
                self.__class__.__name__,
                "(",
                ", ".join(
                    [
                        f"{k}={v:.4f}" if isinstance(v, Number) else f"{k}={v}"
                        for k, v in self.get_params().items()
                    ]
                ),
                ")>",
            ]
        )


class ContinuousDistMixin(DistMixin, metaclass=ABCMeta):
    """
    Mixin class for all continuous probability distributions in wnb.
    """

    _type = "continuous"

    def __init__(self, **kwargs):
        """
        Initializes an instance of the continuous probability distribution with given parameters.
        """
        pass

    def pdf(self, x: float) -> float:
        """Returns the value of probability density function (PDF) at x.

        Args:
            x (float): Input value.

        Returns:
            float: Probability density.
        """
        pass

    @vectorize(signature="(),()->()")
    def __call__(self, x: float) -> float:
        self._check_support(x)
        return self.pdf(x)


class DiscreteDistMixin(DistMixin, metaclass=ABCMeta):
    """
    Mixin class for all discrete probability distributions in wnb.
    """

    _type = "discrete"

    def __init__(self, **kwargs):
        """
        Initializes an instance of the discrete probability distribution with given parameters.
        """
        pass

    def pmf(self, x: float) -> float:
        """Returns the value of probability mass function (PMF) at x.

        Args:
            x (float): Input value.

        Returns:
            float: Probability mass.
        """
        pass

    @vectorize(signature="(),()->()")
    def __call__(self, x: float) -> float:
        self._check_support(x)
        return self.pmf(x)
